
#ifndef _UTIL_UNITS_H
#define _UTIL_UNITS_H

/* useful unit macros for constants */
#define _KB_bytes(x)		(((Bitu)x) << (Bitu)10UL)
#define _MB_bytes(x)		(((Bitu)x) << (Bitu)20UL)
#define _GB_bytes(x)		(((Bitu)x) << (Bitu)30UL)

#endif /* _UTIL_UNITS_H */
