
#include "winres.h"
#include "../include/resource.h"

// icon resource
dosbox_ico ICON "dosbox.ico"

IDR_MENU MENU DISCARDABLE 
BEGIN
	POPUP "Main"
	BEGIN
		MENUITEM "Quick Launch",		ID_OPENFILE
		MENUITEM "&Mapper\tCtrl-F1",		ID_KEYMAP
		MENUITEM SEPARATOR
		MENUITEM "Wait on error",		ID_WAITONERR
		MENUITEM "Avoid usage of symkeys",		ID_USESCANCODES
		MENUITEM "Capture/Release mouse\tCtrl-F10",		ID_CAPMOUSE
		MENUITEM "Autolock mouse",			ID_AUTOLOCK
		MENUITEM "Pause\tAlt-Pause",		ID_PAUSE
		MENUITEM SEPARATOR
		POPUP "Send key"
		    BEGIN
  		    MENUITEM "Send Ctrl-Esc", ID_SEND_CTRL_ESC
  		    MENUITEM "Send Alt-Tab", ID_SEND_ALT_TAB
   		    MENUITEM "Send Ctrl-Alt-Del", ID_SEND_CTRL_ALT_DEL
		    END
		POPUP "Configuration"
		BEGIN
			MENUITEM "Basic settings",		ID_DOSBOX_SECTION
			MENUITEM "Save the current settings",		ID_REFRESH
			MENUITEM "Save the current language strings",		ID_SAVELANG
		END
		MENUITEM "Show &details",		ID_HIDECYCL
		MENUITEM "Show &console",		ID_SHOWCONSOLE
		MENUITEM "&Reset system\tCtrl-Alt-Pause", ID_RESET
		MENUITEM "Restart DOSBox-X\tCtrl-Alt-Home",		ID_RESTART
		MENUITEM "E&xit\tCtrl-F9",		ID_QUIT
	END
	POPUP "Cpu"
	BEGIN
		MENUITEM "Set Cycles",		ID_CYCLE
		MENUITEM SEPARATOR
		MENUITEM "Maximize Cycles",		ID_AUTOCYCLE
		MENUITEM "Automated Cycles Detection",		ID_AUTODETER
		MENUITEM "Turbo\tAlt-F11", ID_CPU_TURBO
		MENUITEM SEPARATOR
		POPUP "Core"
		    BEGIN
    		MENUITEM "Auto",		ID_AUTO
       		MENUITEM "Normal\tCtrl-1",		ID_NORMAL
    		MENUITEM "Full\tCtrl-2",		ID_FULL
    		MENUITEM "Dynamic\tCtrl-3",		ID_DYNAMIC
    		MENUITEM "Simple\tCtrl-4",		ID_SIMPLE
		    END
		POPUP "Type"
		    BEGIN
    		MENUITEM "Auto",		ID_CPUTYPE_AUTO
			MENUITEM "8086", ID_CPUTYPE_8086
			MENUITEM "8086 Prefetch", ID_CPUTYPE_8086_PREFETCH
			MENUITEM "80186", ID_CPUTYPE_80186
			MENUITEM "80186 Prefetch", ID_CPUTYPE_80186_PREFETCH
			MENUITEM "286", ID_CPUTYPE_286
			MENUITEM "286 Prefetch", ID_CPUTYPE_286_PREFETCH
    		MENUITEM "386",		ID_CPUTYPE_386
    		MENUITEM "386 Prefetch",		ID_CPUTYPE_386_PREFETCH
    		MENUITEM "486",		ID_CPUTYPE_486
			MENUITEM "486 Prefetch", ID_CPUTYPE_486_PREFETCH
			MENUITEM "Pentium",		ID_CPUTYPE_PENTIUM
    		MENUITEM "Pentium MMX",		ID_CPUTYPE_PENTIUM_MMX
			MENUITEM "Pentium Pro", ID_CPUTYPE_PENTIUM_PRO
			END
		MENUITEM SEPARATOR
		MENUITEM "Advanced",    ID_CPU_ADVANCED
	END
	POPUP "Video"
	BEGIN
		MENUITEM "Fullscreen\tAlt-Enter",				       ID_FULLSCREEN
		MENUITEM "Hide Menu Bar\tCtrl-Alt-Enter",		ID_TOGGLE
		MENUITEM "Always on top",			ID_ALWAYS_ON_TOP
        MENUITEM "Reset window scale to 100%",                 ID_RESET_RESCALE
		MENUITEM SEPARATOR
		MENUITEM "Aspect Correction",				       ID_ASPECT
		MENUITEM "Double Buffering (Fullscreen)",				       ID_FULLDOUBLE
		MENUITEM SEPARATOR
	POPUP "Frameskip"
	BEGIN
		MENUITEM "0",		ID_SKIP_0
		MENUITEM "1",		ID_SKIP_1
		MENUITEM "2",		ID_SKIP_2
		MENUITEM "3",		ID_SKIP_3
		MENUITEM "4",		ID_SKIP_4
		MENUITEM "5",		ID_SKIP_5
		MENUITEM "6",		ID_SKIP_6
		MENUITEM "7",		ID_SKIP_7
		MENUITEM "8",		ID_SKIP_8
		MENUITEM "9",		ID_SKIP_9
		MENUITEM "10",		ID_SKIP_10
	END
	POPUP "Scaler"
	BEGIN
		MENUITEM "Forced",		ID_FORCESCALER
		MENUITEM SEPARATOR
		MENUITEM "None",				       ID_NONE
		MENUITEM "Normal 2X",				       ID_NORMAL2X
		MENUITEM "Normal 3X",				       ID_NORMAL3X
		MENUITEM "Normal 4X",				       ID_NORMAL4X
		MENUITEM "Normal 5X",				       ID_NORMAL5X
		MENUITEM "Hardware None",				   ID_HARDWARE_NONE
		MENUITEM "Hardware 2X",				       ID_HARDWARE2X
		MENUITEM "Hardware 3X",				       ID_HARDWARE3X
		MENUITEM "Hardware 4X",				       ID_HARDWARE4X
		MENUITEM "Hardware 5X",				       ID_HARDWARE5X
		MENUITEM "TV 2X",				       ID_TV2X
		MENUITEM "TV 3X",				       ID_TV3X
		MENUITEM "Scan 2X",				       ID_SCAN2X
		MENUITEM "Scan 3X",				       ID_SCAN3X
		MENUITEM "RGB 2X",				       ID_RGB2X
		MENUITEM "RGB 3X",				       ID_RGB3X
		MENUITEM "Advanced MAME 2X",				       ID_ADVMAME2X
		MENUITEM "Advanced MAME 3X",				       ID_ADVMAME3X
		MENUITEM "HQ 2X",				       ID_HQ2X
		MENUITEM "HQ 3X",				       ID_HQ3X
		MENUITEM "Advanced Interpolation 2X",				       ID_ADVINTERP2X
		MENUITEM "Advanced Interpolation 3X",				       ID_ADVINTERP3X
		MENUITEM "2xSai",				       ID_2XSAI
		MENUITEM "Super2xSai",				       ID_SUPER2XSAI
		MENUITEM "SuperEagle",				       ID_SUPEREAGLE
		MENUITEM "xBRZ",				       ID_XBRZ
	END
	POPUP "Output"
	BEGIN
		MENUITEM "Surface",				       ID_SURFACE
		MENUITEM "Direct3D",				       ID_DIRECT3D
		MENUITEM "OpenGL",				       ID_OPENGL
		MENUITEM "OpenGL_NB",				       ID_OPENGLNB
	END
	POPUP "Resolution"
	BEGIN
		POPUP "Window"
		BEGIN
		MENUITEM "Original",		ID_WINRES_ORIGINAL
		MENUITEM "Desktop",		ID_WINRES_DESKTOP
		MENUITEM "User-defined",		ID_WINRES_USER
		END
		POPUP "Fullscreen"
		BEGIN
		MENUITEM "Original",		ID_WINFULL_ORIGINAL
		MENUITEM "Desktop",		ID_WINFULL_DESKTOP
		MENUITEM "User-defined",		ID_WINFULL_USER
		END
	END
	POPUP "V-Sync"
	BEGIN
		MENUITEM "On",	ID_VSYNC_ON
		MENUITEM "Force",	ID_VSYNC_FORCE
		MENUITEM "Host",	ID_VSYNC_HOST
		MENUITEM "Off",	ID_VSYNC_OFF
		MENUITEM SEPARATOR
		MENUITEM "Set syncrate",	ID_VSYNC
	END
	POPUP "Overscan"
	BEGIN
		MENUITEM "0",	ID_OVERSCAN_0
		MENUITEM "1",	ID_OVERSCAN_1
		MENUITEM "2",	ID_OVERSCAN_2
		MENUITEM "3",	ID_OVERSCAN_3
		MENUITEM "4",	ID_OVERSCAN_4
		MENUITEM "5",	ID_OVERSCAN_5
		MENUITEM "6",	ID_OVERSCAN_6
		MENUITEM "7",	ID_OVERSCAN_7
		MENUITEM "8",	ID_OVERSCAN_8
		MENUITEM "9",	ID_OVERSCAN_9
		MENUITEM "10",	ID_OVERSCAN_10
	END
	POPUP "Compatibility"
	BEGIN
		MENUITEM "Allow 9-pixel wide text mode fonts",	ID_CHAR9
		MENUITEM "Multiscan",	ID_MULTISCAN
	END
	POPUP "PC-98"
	BEGIN
		MENUITEM "5MHz GDC clock", ID_PC98_GDC5MHZ
		MENUITEM "Allow 200-line scanline effect", ID_PC98_200SCANLINEEFFECT
		MENUITEM "Allow 4 display partitions in graphics layer", ID_PC98_FOURPARTITIONSGRAPHICS
		MENUITEM SEPARATOR
		MENUITEM "Enable EGC", ID_PC98_ENABLEEGC
		MENUITEM "Enable GRCG", ID_PC98_ENABLEGRCG
		MENUITEM "Enable analog display", ID_PC98_ENABLE16COLORS
		MENUITEM SEPARATOR
		MENUITEM "Clear text layer", ID_PC98_CLEAR_TEXT_LAYER
		MENUITEM "Clear graphics layer", ID_PC98_CLEAR_GRAPHICS_LAYER
	END
	MENUITEM SEPARATOR
		MENUITEM "Direct3D Pixelshader",	ID_D3D_PS
	END
	POPUP "Sound"
	BEGIN
		MENUITEM "Mixer", ID_MIXER_SECTION
		MENUITEM "Mute", ID_MUTE
		MENUITEM "Swap stereo", ID_SWAPSTEREO
	    MENUITEM SEPARATOR
		POPUP "MIDI"
		BEGIN
			POPUP "MPU401"
			BEGIN
			MENUITEM "None", ID_MIDI_NONE
			MENUITEM "UART", ID_MIDI_UART
			MENUITEM "Intelligent", ID_MIDI_INTELLI
			END
			POPUP "Device"
			BEGIN
			MENUITEM "None", ID_MIDI_DEV_NONE
			MENUITEM "Default", ID_MIDI_DEFAULT
			MENUITEM "ALSA", ID_MIDI_ALSA
			MENUITEM "OSS", ID_MIDI_OSS
			MENUITEM "Win32", ID_MIDI_WIN32
			MENUITEM "Core audio", ID_MIDI_COREAUDIO
			MENUITEM "MT-32", ID_MIDI_MT32
			MENUITEM "Fluidsynth", ID_MIDI_SYNTH
			MENUITEM "Timidity server", ID_MIDI_TIMIDITY
			END
			MENUITEM SEPARATOR
			POPUP "MT-32 reverb mode"
			BEGIN
			MENUITEM "Auto", ID_MIDI_MT32_REVERBMODE_AUTO
			MENUITEM "0", ID_MIDI_MT32_REVERBMODE_0
			MENUITEM "1", ID_MIDI_MT32_REVERBMODE_1
			MENUITEM "2", ID_MIDI_MT32_REVERBMODE_2
			MENUITEM "3", ID_MIDI_MT32_REVERBMODE_3
		    END
			POPUP "MT-32 reverb time"
			BEGIN
			MENUITEM "0", ID_MIDI_MT32_REVERBTIME_0
			MENUITEM "1", ID_MIDI_MT32_REVERBTIME_1
			MENUITEM "2", ID_MIDI_MT32_REVERBTIME_2
			MENUITEM "3", ID_MIDI_MT32_REVERBTIME_3
			MENUITEM "4", ID_MIDI_MT32_REVERBTIME_4
			MENUITEM "5", ID_MIDI_MT32_REVERBTIME_5
			MENUITEM "6", ID_MIDI_MT32_REVERBTIME_6
			MENUITEM "7", ID_MIDI_MT32_REVERBTIME_7
			END
			POPUP "MT-32 reverb level"
			BEGIN
			MENUITEM "0", ID_MIDI_MT32_REVERBLEV_0
			MENUITEM "1", ID_MIDI_MT32_REVERBLEV_1
			MENUITEM "2", ID_MIDI_MT32_REVERBLEV_2
			MENUITEM "3", ID_MIDI_MT32_REVERBLEV_3
			MENUITEM "4", ID_MIDI_MT32_REVERBLEV_4
			MENUITEM "5", ID_MIDI_MT32_REVERBLEV_5
			MENUITEM "6", ID_MIDI_MT32_REVERBLEV_6
			MENUITEM "7", ID_MIDI_MT32_REVERBLEV_7
			END
			POPUP "MT-32 DAC modes"
			BEGIN
			MENUITEM "Auto", ID_MIDI_MT32_DAC_AUTO
			MENUITEM "0", ID_MIDI_MT32_DAC_0
			MENUITEM "1", ID_MIDI_MT32_DAC_1
			MENUITEM "2", ID_MIDI_MT32_DAC_2
			MENUITEM "3", ID_MIDI_MT32_DAC_3
			END
			POPUP "MT-32 Reverse Stereo"
			BEGIN
			MENUITEM "On", ID_MIDI_MT32_REVERSESTEREO_TRUE
			MENUITEM "Off", ID_MIDI_MT32_REVERSESTEREO_FALSE
			END
			MENUITEM SEPARATOR
			MENUITEM "Advanced", ID_MIDI_ADVANCED
		END
		POPUP "Sound Blaster"
		BEGIN
			POPUP "Type"
			BEGIN
			MENUITEM "None", ID_SB_NONE
			MENUITEM "Sound Blaster 1", ID_SB_SB1
			MENUITEM "Sound Blaster 2", ID_SB_SB2
			MENUITEM "Sound Blaster Pro 1", ID_SB_SBPRO1
			MENUITEM "Sound Blaster Pro 2", ID_SB_SBPRO2
			MENUITEM "Sound Blaster 16", ID_SB_SB16
			MENUITEM "Sound Blaster 16 Vibra", ID_SB_SB16VIBRA
			MENUITEM "Gameblaster", ID_SB_GB
			END
			POPUP "Base"
			BEGIN
			MENUITEM "220", ID_SB_220
			MENUITEM "240", ID_SB_240
			MENUITEM "260", ID_SB_260
			MENUITEM "280", ID_SB_280
			MENUITEM "2a0", ID_SB_2a0
			MENUITEM "2c0", ID_SB_2c0
			MENUITEM "2e0", ID_SB_2e0
			MENUITEM "300", ID_SB_300
			END
			POPUP "Hardware Base"
			BEGIN
			MENUITEM "210", ID_SB_HW210
			MENUITEM "220", ID_SB_HW220
			MENUITEM "230", ID_SB_HW230
			MENUITEM "240", ID_SB_HW240
			MENUITEM "250", ID_SB_HW250
			MENUITEM "260", ID_SB_HW260
			MENUITEM "280", ID_SB_HW280
		    END
			POPUP "IRQ"
			BEGIN
			MENUITEM "3", ID_SB_IRQ_3
			MENUITEM "5", ID_SB_IRQ_5
			MENUITEM "7", ID_SB_IRQ_7
			MENUITEM "9", ID_SB_IRQ_9
			MENUITEM "10", ID_SB_IRQ_10
			MENUITEM "11", ID_SB_IRQ_11
			MENUITEM "12", ID_SB_IRQ_12
			END
			POPUP "DMA"
			BEGIN
			MENUITEM "0", ID_SB_DMA_0
			MENUITEM "1", ID_SB_DMA_1
			MENUITEM "3", ID_SB_DMA_3
			MENUITEM "5", ID_SB_DMA_5
			MENUITEM "6", ID_SB_DMA_6
			MENUITEM "7", ID_SB_DMA_7
			END
			POPUP "HDMA"
			BEGIN
			MENUITEM "0", ID_SB_HDMA_0
			MENUITEM "1", ID_SB_HDMA_1
			MENUITEM "3", ID_SB_HDMA_3
			MENUITEM "5", ID_SB_HDMA_5
			MENUITEM "6", ID_SB_HDMA_6
			MENUITEM "7", ID_SB_HDMA_7
			END
			POPUP "OPL Mode"
			BEGIN
			MENUITEM "Auto", ID_SB_OPL_AUTO
			MENUITEM "None", ID_SB_OPL_NONE
			MENUITEM "CMS", ID_SB_OPL_CMS
			MENUITEM "OPL2", ID_SB_OPL_OPL2
			MENUITEM "Dual OPL2", ID_SB_OPL_DUALOPL2
			MENUITEM "OPL3", ID_SB_OPL_OPL3
			MENUITEM "HARDWARE", ID_SB_OPL_HARDWARE
			MENUITEM "HARDWAREGB", ID_SB_OPL_HARDWAREGB
			END
			POPUP "OPL Rate"
			BEGIN
			MENUITEM "49716",  ID_SB_OPL_49716
			MENUITEM "48000",  ID_SB_OPL_48000
			MENUITEM "44100",  ID_SB_OPL_44100
			MENUITEM "32000",  ID_SB_OPL_32000
			MENUITEM "22050",  ID_SB_OPL_22050
			MENUITEM "16000",  ID_SB_OPL_16000
			MENUITEM "11025",  ID_SB_OPL_11025
			MENUITEM "8000",  ID_SB_OPL_8000
			END
			POPUP "OPL Emulation"
			BEGIN
			MENUITEM "Default",  ID_SB_OPL_EMU_DEFAULT
			MENUITEM "Compatible",  ID_SB_OPL_EMU_COMPAT
			MENUITEM "Fast",  ID_SB_OPL_EMU_FAST
			END
		END
		POPUP "Gravis Ultrasound"
		BEGIN
		MENUITEM "Enable", ID_GUS_TRUE
			POPUP "Rate"
			BEGIN
				MENUITEM "49716", ID_GUS_49716
				MENUITEM "48000", ID_GUS_48000
				MENUITEM "44100", ID_GUS_44100
				MENUITEM "32000", ID_GUS_32000
				MENUITEM "22050", ID_GUS_22050
				MENUITEM "16000", ID_GUS_16000
				MENUITEM "11025", ID_GUS_11025
				MENUITEM "8000", ID_GUS_8000
			END
			POPUP "Base"
			BEGIN
				MENUITEM "300",ID_GUS_300
				MENUITEM "220",ID_GUS_220
				MENUITEM "240",ID_GUS_240
				MENUITEM "260",ID_GUS_260
				MENUITEM "280",ID_GUS_280
				MENUITEM "2a0",ID_GUS_2a0
				MENUITEM "2c0",ID_GUS_2c0
				MENUITEM "2e0",ID_GUS_2e0
			END
			POPUP "IRQ"
			BEGIN
				MENUITEM "3", ID_GUS_IRQ_3
				MENUITEM "5", ID_GUS_IRQ_5
				MENUITEM "7", ID_GUS_IRQ_7
				MENUITEM "9", ID_GUS_IRQ_9
				MENUITEM "10", ID_GUS_IRQ_10
				MENUITEM "11", ID_GUS_IRQ_11
				MENUITEM "12", ID_GUS_IRQ_12
			END
			POPUP "DMA"
			BEGIN
				MENUITEM "0", ID_GUS_DMA_0
				MENUITEM "1", ID_GUS_DMA_1
				MENUITEM "3", ID_GUS_DMA_3
				MENUITEM "5", ID_GUS_DMA_5
				MENUITEM "6", ID_GUS_DMA_6
				MENUITEM "7", ID_GUS_DMA_7
			END
		END
		POPUP "Innovation SSI-2001"
		BEGIN
		MENUITEM "Enable", ID_INNOVA_TRUE
			POPUP "Rate"
			BEGIN
				MENUITEM "49716",  ID_INNOVA_49716
				MENUITEM "48000",  ID_INNOVA_48000
				MENUITEM "44100",  ID_INNOVA_44100
				MENUITEM "32000",  ID_INNOVA_32000
				MENUITEM "16000",  ID_INNOVA_16000
				MENUITEM "22050",  ID_INNOVA_22050
				MENUITEM "11025",  ID_INNOVA_11025
				MENUITEM "8000",  ID_INNOVA_8000
			END
			POPUP "SID Base"
			BEGIN
				MENUITEM "300",  ID_INNOVA_300
				MENUITEM "280",  ID_INNOVA_280
				MENUITEM "260",  ID_INNOVA_260
				MENUITEM "240",  ID_INNOVA_240
				MENUITEM "220",  ID_INNOVA_220
				MENUITEM "2A0",  ID_INNOVA_2A0
				MENUITEM "2C0",  ID_INNOVA_2C0
				MENUITEM "2E0",  ID_INNOVA_2E0
			END
			POPUP "Quality"
			BEGIN
				MENUITEM "3",  ID_INNOVA_3
				MENUITEM "2",  ID_INNOVA_2
				MENUITEM "1",  ID_INNOVA_1
				MENUITEM "0",  ID_INNOVA_0
			END
		END
		POPUP "PC Speaker"
		BEGIN
		MENUITEM "Enable", ID_PCSPEAKER_TRUE
			POPUP "Rate"
			BEGIN
				MENUITEM "49716",  ID_PCSPEAKER_49716
				MENUITEM "48000",  ID_PCSPEAKER_48000
				MENUITEM "44100",  ID_PCSPEAKER_44100
				MENUITEM "32000",  ID_PCSPEAKER_32000
				MENUITEM "22050",  ID_PCSPEAKER_22050
				MENUITEM "16000",  ID_PCSPEAKER_16000
				MENUITEM "11025",  ID_PCSPEAKER_11025
				MENUITEM "8000",  ID_PCSPEAKER_8000
			END
			POPUP "Tandy"
			BEGIN
				MENUITEM "On",  ID_TANDY_ON
				MENUITEM "Off",  ID_TANDY_OFF
				MENUITEM "Auto",  ID_TANDY_AUTO
				POPUP "Rate"
				BEGIN
				MENUITEM "49716",ID_TANDY_49716
				MENUITEM "48000",ID_TANDY_48000
				MENUITEM "44100",ID_TANDY_44100
				MENUITEM "32000",ID_TANDY_32000
				MENUITEM "22050",ID_TANDY_22050
				MENUITEM "16000",ID_TANDY_16000
				MENUITEM "11025",ID_TANDY_11025
				MENUITEM "8000",ID_TANDY_8000
				END
			END
			POPUP "Disney"
			BEGIN
				MENUITEM "Enable", ID_DISNEY_TRUE
				MENUITEM "Disable", ID_DISNEY_FALSE
			END
			POPUP "PS/1 Audio"
			BEGIN
				MENUITEM "On", ID_PS1_ON
				MENUITEM "Off", ID_PS1_OFF
				POPUP "Rate"
				BEGIN
				MENUITEM "49716",ID_PS1_49716
				MENUITEM "48000",ID_PS1_48000
				MENUITEM "44100",ID_PS1_44100
				MENUITEM "32000",ID_PS1_32000
				MENUITEM "22050",ID_PS1_22050
				MENUITEM "16000",ID_PS1_16000
				MENUITEM "11025",ID_PS1_11025
				MENUITEM "8000",ID_PS1_8000
				END
			END
		END
	END
	POPUP "Dos"
	BEGIN
		POPUP "Memory"
		BEGIN
		MENUITEM "XMS",				       ID_XMS
		POPUP "EMS"
			BEGIN
			MENUITEM "True",				       ID_EMS_TRUE
			MENUITEM "False",				       ID_EMS_FALSE
			MENUITEM "EMS Board",				       ID_EMS_EMSBOARD
			MENUITEM "EMM386",				       ID_EMS_EMM386
			END
		MENUITEM "UMB",				       ID_UMB
		END
	POPUP "Storage"
	BEGIN
		MENUITEM "Set Default Local Free Size",	ID_HDD_SIZE
		MENUITEM "Enable Automatic Mounting",	ID_AUTOMOUNT
	END
	MENUITEM SEPARATOR
    POPUP "Mouse"
    BEGIN
		MENUITEM "Internal Emulation",	ID_MOUSE
		MENUITEM "Y-axis Reverse",	ID_MOUSE_VERTICAL
		MENUITEM SEPARATOR
		MENUITEM "Sensitivity", ID_MOUSE_SENSITIVITY
	END
	POPUP "Keyboard"
	BEGIN
		POPUP "Rate/Delay"
		BEGIN
			POPUP "Delay 1"
			BEGIN
				MENUITEM "Rate 1", ID_RATE_1_DELAY_1
				MENUITEM "Rate 2", ID_RATE_2_DELAY_1
				MENUITEM "Rate 3", ID_RATE_3_DELAY_1
				MENUITEM "Rate 4", ID_RATE_4_DELAY_1
				MENUITEM "Rate 5", ID_RATE_5_DELAY_1
				MENUITEM "Rate 6", ID_RATE_6_DELAY_1
				MENUITEM "Rate 7", ID_RATE_7_DELAY_1
				MENUITEM "Rate 8", ID_RATE_8_DELAY_1
				MENUITEM "Rate 9", ID_RATE_9_DELAY_1
				MENUITEM "Rate 10", ID_RATE_10_DELAY_1
				MENUITEM "Rate 11", ID_RATE_11_DELAY_1
				MENUITEM "Rate 12", ID_RATE_12_DELAY_1
				MENUITEM "Rate 13", ID_RATE_13_DELAY_1
				MENUITEM "Rate 14", ID_RATE_14_DELAY_1
				MENUITEM "Rate 15", ID_RATE_15_DELAY_1
				MENUITEM "Rate 16", ID_RATE_16_DELAY_1
				MENUITEM "Rate 17", ID_RATE_17_DELAY_1
				MENUITEM "Rate 18", ID_RATE_18_DELAY_1
				MENUITEM "Rate 19", ID_RATE_19_DELAY_1
				MENUITEM "Rate 20", ID_RATE_20_DELAY_1
				MENUITEM "Rate 21", ID_RATE_21_DELAY_1
				MENUITEM "Rate 22", ID_RATE_22_DELAY_1
				MENUITEM "Rate 23", ID_RATE_23_DELAY_1
				MENUITEM "Rate 24", ID_RATE_24_DELAY_1
				MENUITEM "Rate 25", ID_RATE_25_DELAY_1
				MENUITEM "Rate 26", ID_RATE_26_DELAY_1
				MENUITEM "Rate 27", ID_RATE_27_DELAY_1
				MENUITEM "Rate 28", ID_RATE_28_DELAY_1
				MENUITEM "Rate 29", ID_RATE_29_DELAY_1
				MENUITEM "Rate 30", ID_RATE_30_DELAY_1
				MENUITEM "Rate 31", ID_RATE_31_DELAY_1
				MENUITEM "Rate 32", ID_RATE_32_DELAY_1
			END
			POPUP "Delay 2"
			BEGIN
				MENUITEM "Rate 1", ID_RATE_1_DELAY_2
				MENUITEM "Rate 2", ID_RATE_2_DELAY_2
				MENUITEM "Rate 3", ID_RATE_3_DELAY_2
				MENUITEM "Rate 4", ID_RATE_4_DELAY_2
				MENUITEM "Rate 5", ID_RATE_5_DELAY_2
				MENUITEM "Rate 6", ID_RATE_6_DELAY_2
				MENUITEM "Rate 7", ID_RATE_7_DELAY_2
				MENUITEM "Rate 8", ID_RATE_8_DELAY_2
				MENUITEM "Rate 9", ID_RATE_9_DELAY_2
				MENUITEM "Rate 10", ID_RATE_10_DELAY_2
				MENUITEM "Rate 11", ID_RATE_11_DELAY_2
				MENUITEM "Rate 12", ID_RATE_12_DELAY_2
				MENUITEM "Rate 13", ID_RATE_13_DELAY_2
				MENUITEM "Rate 14", ID_RATE_14_DELAY_2
				MENUITEM "Rate 15", ID_RATE_15_DELAY_2
				MENUITEM "Rate 16", ID_RATE_16_DELAY_2
				MENUITEM "Rate 17", ID_RATE_17_DELAY_2
				MENUITEM "Rate 18", ID_RATE_18_DELAY_2
				MENUITEM "Rate 19", ID_RATE_19_DELAY_2
				MENUITEM "Rate 20", ID_RATE_20_DELAY_2
				MENUITEM "Rate 21", ID_RATE_21_DELAY_2
				MENUITEM "Rate 22", ID_RATE_22_DELAY_2
				MENUITEM "Rate 23", ID_RATE_23_DELAY_2
				MENUITEM "Rate 24", ID_RATE_24_DELAY_2
				MENUITEM "Rate 25", ID_RATE_25_DELAY_2
				MENUITEM "Rate 26", ID_RATE_26_DELAY_2
				MENUITEM "Rate 27", ID_RATE_27_DELAY_2
				MENUITEM "Rate 28", ID_RATE_28_DELAY_2
				MENUITEM "Rate 29", ID_RATE_29_DELAY_2
				MENUITEM "Rate 30", ID_RATE_30_DELAY_2
				MENUITEM "Rate 31", ID_RATE_31_DELAY_2
				MENUITEM "Rate 32", ID_RATE_32_DELAY_2
			END
			POPUP "Delay 3"
			BEGIN
				MENUITEM "Rate 1", ID_RATE_1_DELAY_3
				MENUITEM "Rate 2", ID_RATE_2_DELAY_3
				MENUITEM "Rate 3", ID_RATE_3_DELAY_3
				MENUITEM "Rate 4", ID_RATE_4_DELAY_3
				MENUITEM "Rate 5", ID_RATE_5_DELAY_3
				MENUITEM "Rate 6", ID_RATE_6_DELAY_3
				MENUITEM "Rate 7", ID_RATE_7_DELAY_3
				MENUITEM "Rate 8", ID_RATE_8_DELAY_3
				MENUITEM "Rate 9", ID_RATE_9_DELAY_3
				MENUITEM "Rate 10", ID_RATE_10_DELAY_3
				MENUITEM "Rate 11", ID_RATE_11_DELAY_3
				MENUITEM "Rate 12", ID_RATE_12_DELAY_3
				MENUITEM "Rate 13", ID_RATE_13_DELAY_3
				MENUITEM "Rate 14", ID_RATE_14_DELAY_3
				MENUITEM "Rate 15", ID_RATE_15_DELAY_3
				MENUITEM "Rate 16", ID_RATE_16_DELAY_3
				MENUITEM "Rate 17", ID_RATE_17_DELAY_3
				MENUITEM "Rate 18", ID_RATE_18_DELAY_3
				MENUITEM "Rate 19", ID_RATE_19_DELAY_3
				MENUITEM "Rate 20", ID_RATE_20_DELAY_3
				MENUITEM "Rate 21", ID_RATE_21_DELAY_3
				MENUITEM "Rate 22", ID_RATE_22_DELAY_3
				MENUITEM "Rate 23", ID_RATE_23_DELAY_3
				MENUITEM "Rate 24", ID_RATE_24_DELAY_3
				MENUITEM "Rate 25", ID_RATE_25_DELAY_3
				MENUITEM "Rate 26", ID_RATE_26_DELAY_3
				MENUITEM "Rate 27", ID_RATE_27_DELAY_3
				MENUITEM "Rate 28", ID_RATE_28_DELAY_3
				MENUITEM "Rate 29", ID_RATE_29_DELAY_3
				MENUITEM "Rate 30", ID_RATE_30_DELAY_3
				MENUITEM "Rate 31", ID_RATE_31_DELAY_3
				MENUITEM "Rate 32", ID_RATE_32_DELAY_3
			END
			POPUP "Delay 4"
			BEGIN
				MENUITEM "Rate 1", ID_RATE_1_DELAY_4
				MENUITEM "Rate 2", ID_RATE_2_DELAY_4
				MENUITEM "Rate 3", ID_RATE_3_DELAY_4
				MENUITEM "Rate 4", ID_RATE_4_DELAY_4
				MENUITEM "Rate 5", ID_RATE_5_DELAY_4
				MENUITEM "Rate 6", ID_RATE_6_DELAY_4
				MENUITEM "Rate 7", ID_RATE_7_DELAY_4
				MENUITEM "Rate 8", ID_RATE_8_DELAY_4
				MENUITEM "Rate 9", ID_RATE_9_DELAY_4
				MENUITEM "Rate 10", ID_RATE_10_DELAY_4
				MENUITEM "Rate 11", ID_RATE_11_DELAY_4
				MENUITEM "Rate 12", ID_RATE_12_DELAY_4
				MENUITEM "Rate 13", ID_RATE_13_DELAY_4
				MENUITEM "Rate 14", ID_RATE_14_DELAY_4
				MENUITEM "Rate 15", ID_RATE_15_DELAY_4
				MENUITEM "Rate 16", ID_RATE_16_DELAY_4
				MENUITEM "Rate 17", ID_RATE_17_DELAY_4
				MENUITEM "Rate 18", ID_RATE_18_DELAY_4
				MENUITEM "Rate 19", ID_RATE_19_DELAY_4
				MENUITEM "Rate 20", ID_RATE_20_DELAY_4
				MENUITEM "Rate 21", ID_RATE_21_DELAY_4
				MENUITEM "Rate 22", ID_RATE_22_DELAY_4
				MENUITEM "Rate 23", ID_RATE_23_DELAY_4
				MENUITEM "Rate 24", ID_RATE_24_DELAY_4
				MENUITEM "Rate 25", ID_RATE_25_DELAY_4
				MENUITEM "Rate 26", ID_RATE_26_DELAY_4
				MENUITEM "Rate 27", ID_RATE_27_DELAY_4
				MENUITEM "Rate 28", ID_RATE_28_DELAY_4
				MENUITEM "Rate 29", ID_RATE_29_DELAY_4
				MENUITEM "Rate 30", ID_RATE_30_DELAY_4
				MENUITEM "Rate 31", ID_RATE_31_DELAY_4
				MENUITEM "Rate 32", ID_RATE_32_DELAY_4
			END
		END
		POPUP "Layout"
		BEGIN
			MENUITEM "Auto",	ID_KEY_NONE
			MENUITEM "BG - Bulgaria",	ID_KEY_BG
			MENUITEM "CZ - Czech Republic",	ID_KEY_CZ
			MENUITEM "FR - France",	ID_KEY_FR
			MENUITEM "GK - Greece",	ID_KEY_GK
			MENUITEM "GR - Germany",	ID_KEY_GR
			MENUITEM "HR - Croatia",	ID_KEY_HR
			MENUITEM "HU - Hungary",	ID_KEY_HU
			MENUITEM "IT - Italy",	ID_KEY_IT
			MENUITEM "NL - Netherlands",	ID_KEY_NL
			MENUITEM "NO - Norway",	ID_KEY_NO
			MENUITEM "PL - Poland",	ID_KEY_PL
			MENUITEM "RU - Russian Federation",	ID_KEY_RU
			MENUITEM "SK - Slovakia",	ID_KEY_SK
			MENUITEM "SP - Spain",	ID_KEY_SP
			MENUITEM "SU - Finland",	ID_KEY_SU
			MENUITEM "SV - Sweden",	ID_KEY_SV
			MENUITEM "BE - Belgium",	 ID_KEY_BE
			MENUITEM "BR - Brazil",	 ID_KEY_BR
			MENUITEM "CF - Canadian (French)",	 ID_KEY_CF
			POPUP "More"
			BEGIN
				MENUITEM "DK - Denmark",	 ID_KEY_DK
				MENUITEM "LA - Latin America",	 ID_KEY_LA
				MENUITEM "PO - Portugal",	 ID_KEY_PO
				MENUITEM "SF - Switzerland (French)",	 ID_KEY_SF
				MENUITEM "SG - Switzerland (German)",	 ID_KEY_SG
				MENUITEM "UK - United Kingdom (English)",	 ID_KEY_UK
				MENUITEM "US - United States (English)",	 ID_KEY_US
				MENUITEM "YU - Yugoslavia",	 ID_KEY_YU
				MENUITEM "FO - Unknown",	 ID_KEY_FO
				MENUITEM "MK - Unknown",	 ID_KEY_MK
				MENUITEM "MT - Unknown",	 ID_KEY_MT
				MENUITEM "PH - Unknown",	 ID_KEY_PH
				MENUITEM "RO - Unknown",	 ID_KEY_RO
				MENUITEM "SQ - Unknown",	 ID_KEY_SQ
				MENUITEM "TM - Unknown",	 ID_KEY_TM
				MENUITEM "TR - Unknown",	 ID_KEY_TR
				MENUITEM "UX - Unknown",	 ID_KEY_UX
				MENUITEM "YC - Unknown",	 ID_KEY_YC
				MENUITEM "DV - Two-handed DVORAK",	 ID_KEY_DV
				MENUITEM "RH - Right-handed DVORAK",	 ID_KEY_RH
				MENUITEM "LH - Left-handed DVORAK",	 ID_KEY_LH
			END
		END
	END
	POPUP "Joystick"
	BEGIN
		MENUITEM "Auto",    ID_JOYSTICKTYPE_AUTO
	    MENUITEM "None",    ID_JOYSTICKTYPE_NONE
	    MENUITEM "2 Axis",    ID_JOYSTICKTYPE_2AXIS
	    MENUITEM "4 Axis",    ID_JOYSTICKTYPE_4AXIS
	    MENUITEM "4 Axis 2",    ID_JOYSTICKTYPE_4AXIS_2
	    MENUITEM "FCS",    ID_JOYSTICKTYPE_FCS
	    MENUITEM "CH",    ID_JOYSTICKTYPE_CH
        MENUITEM SEPARATOR
	    MENUITEM "Timed",    ID_JOYSTICK_TIMED
	    MENUITEM "Auto fire",    ID_JOYSTICK_AUTOFIRE
	    MENUITEM "Swap 34",    ID_JOYSTICK_SWAP34
	    MENUITEM "Button wrap",    ID_JOYSTICK_BUTTONWRAP
	END
	MENUITEM SEPARATOR
	POPUP "Network"
	BEGIN
		MENUITEM "IPX",				       ID_IPXNET
		MENUITEM "NE2000",				       ID_NE2000_SECTION
	END
	MENUITEM "Serial",				       ID_SERIAL_SECTION
	MENUITEM "Parallel / Dongle",				       ID_PARALLEL_SECTION
	MENUITEM "Printer",				       ID_PRINTER_SECTION
    POPUP "Glide Emulation"
    BEGIN
		POPUP "Wrapper"
		BEGIN
			MENUITEM "Enable",	ID_GLIDE_TRUE
			POPUP "LFB"
			BEGIN
				MENUITEM "Full",	ID_GLIDE_LFB_FULL
				MENUITEM "Full_noaux",	ID_GLIDE_LFB_FULL_NOAUX
				MENUITEM "Read",	ID_GLIDE_LFB_READ
				MENUITEM "Read_noaux",	ID_GLIDE_LFB_READ_NOAUX
				MENUITEM "Write",	ID_GLIDE_LFB_WRITE
				MENUITEM "Write_noaux",	ID_GLIDE_LFB_WRITE_NOAUX
				MENUITEM "None",	ID_GLIDE_LFB_NONE
			END
			MENUITEM "Show 3dfx splash screen",	ID_GLIDE_SPLASH
		END
		POPUP "Internal Emulation"
		BEGIN
			MENUITEM "Enable",	ID_GLIDE_EMU
			POPUP "Output"
			BEGIN
				MENUITEM "False",	ID_GLIDE_EMU_FALSE
				MENUITEM "Software",	ID_GLIDE_EMU_SOFTWARE
				MENUITEM "OpenGL",	ID_GLIDE_EMU_OPENGL
				MENUITEM "Auto",	ID_GLIDE_EMU_AUTO
			END
		END
	END
	POPUP "PC-98 PIT master clock"
	BEGIN
		MENUITEM "4MHz/8MHz", ID_PC98_4MHZ_TIMER
		MENUITEM "5MHz/10MHz", ID_PC98_5MHZ_TIMER
	END
	MENUITEM SEPARATOR
	MENUITEM "Restart DOS", ID_RESTART_DOS
	MENUITEM SEPARATOR
	MENUITEM "Edit AUTOEXEC.BAT",			   ID_AUTOEXEC
	MENUITEM "Advanced",				   ID_DOS_ADVANCED
	END
	POPUP "Capture"
	BEGIN
	MENUITEM "Screenshot\tCtrl-F5", ID_SSHOT
	MENUITEM "Movie\tCtrl-Alt-F5",				       ID_MOVIE
	MENUITEM "Wave\tCtrl-F6",				       ID_WAVE
	MENUITEM "Multi-track Wave (AVI)\tCtrl-F7", ID_MTWAVE
	MENUITEM "OPL commands\tCtrl-Alt-F7",				       ID_OPL
	MENUITEM "Raw MIDI commands\tCtrl-Alt-F8",				       ID_MIDI
	END
	POPUP "Drive"
	BEGIN
		POPUP "A"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_A
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_A
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_A
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_A
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_A
			MENUITEM "Unmount",				       ID_UMOUNT_A
			MENUITEM SEPARATOR
			POPUP "Boot from Floppy Image"
		BEGIN
			MENUITEM "Specify Bootable Floppy Image",				       ID_BOOT_A
			MENUITEM "Use Mounted Floppy Image",				       ID_BOOT_A_MOUNTED
		END
		END
		POPUP "B"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_B
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_B
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_B
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_B
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_B
			MENUITEM "Unmount",				       ID_UMOUNT_B
		END
		POPUP "C"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_C
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_C
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_C
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_C
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_C
			MENUITEM "Unmount",				       ID_UMOUNT_C
			MENUITEM SEPARATOR
			POPUP "Boot from Hard Disk Image"
		BEGIN
			MENUITEM "Specify Bootable Hard Disk Image",				       ID_BOOT_C
			MENUITEM "Use Mounted Hard Disk Image",				       ID_BOOT_C_MOUNTED
		END
		END
		POPUP "D"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_D
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_D
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_D
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_D
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_D
			MENUITEM "Unmount",				       ID_UMOUNT_D
			MENUITEM SEPARATOR
			POPUP "Boot from Hard Disk Image"
		BEGIN
			MENUITEM "Specify Bootable Hard Disk Image",				       ID_BOOT_D
			MENUITEM "Use Mounted Hard Disk Image",				       ID_BOOT_D_MOUNTED
		END
		END
		POPUP "E"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_E
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_E
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_E
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_E
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_E
			MENUITEM "Unmount",				       ID_UMOUNT_E
		END
		POPUP "F"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_F
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_F
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_F
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_F
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_F
			MENUITEM "Unmount",				       ID_UMOUNT_F
		END
		POPUP "G"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_G
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_G
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_G
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_G
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_G
			MENUITEM "Unmount",				       ID_UMOUNT_G
		END
		POPUP "H"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_H
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_H
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_H
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_H
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_H
			MENUITEM "Unmount",				       ID_UMOUNT_H
		END
		POPUP "I"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_I
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_I
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_I
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_I
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_I
			MENUITEM "Unmount",				       ID_UMOUNT_I
		END
		POPUP "J"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_J
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_J
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_J
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_J
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_J
			MENUITEM "Unmount",				       ID_UMOUNT_J
		END
		POPUP "K"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_K
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_K
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_K
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_K
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_K
			MENUITEM "Unmount",				       ID_UMOUNT_K
		END
		POPUP "L"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_L
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_L
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_L
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_L
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_L
			MENUITEM "Unmount",				       ID_UMOUNT_L
		END
		POPUP "M"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_M
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_M
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_M
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_M
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_M
			MENUITEM "Unmount",				       ID_UMOUNT_M
		END
		POPUP "N"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_N
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_N
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_N
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_N
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_N
			MENUITEM "Unmount",				       ID_UMOUNT_N
		END
		POPUP "O"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_O
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_O
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_O
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_O
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_O
			MENUITEM "Unmount",				       ID_UMOUNT_O
		END
		POPUP "P"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_P
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_P
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_P
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_P
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_P
			MENUITEM "Unmount",				       ID_UMOUNT_P
		END
		POPUP "Q"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_Q
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_Q
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_Q
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_Q
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_Q
			MENUITEM "Unmount",				       ID_UMOUNT_Q
		END
		POPUP "R"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_R
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_R
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_R
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_R
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_R
			MENUITEM "Unmount",				       ID_UMOUNT_R
		END
		POPUP "S"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_S
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_S
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_S
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_S
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_S
			MENUITEM "Unmount",				       ID_UMOUNT_S
		END
		POPUP "T"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_T
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_T
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_T
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_T
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_T
			MENUITEM "Unmount",				       ID_UMOUNT_T
		END
		POPUP "U"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_U
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_U
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_U
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_U
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_U
			MENUITEM "Unmount",				       ID_UMOUNT_U
		END
		POPUP "V"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_V
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_V
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_V
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_V
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_V
			MENUITEM "Unmount",				       ID_UMOUNT_V
		END
		POPUP "W"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_W
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_W
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_W
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_W
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_W
			MENUITEM "Unmount",				       ID_UMOUNT_W
		END
		POPUP "X"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_X
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_X
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_X
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_X
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_X
			MENUITEM "Unmount",				       ID_UMOUNT_X
		END
		POPUP "Y"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_Y
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_Y
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_Y
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_Y
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_Y
			MENUITEM "Unmount",				       ID_UMOUNT_Y
		END
		POPUP "Z"
		BEGIN
			MENUITEM "Mount Automatically",				       ID_AUTOMOUNT_Z
			MENUITEM "Mount as Hard Disk",				       ID_MOUNT_LOCAL_Z
			MENUITEM "Mount as CD-ROM",				       ID_MOUNT_CDROM_Z
			MENUITEM "Mount as Floppy",				       ID_MOUNT_FLOPPY_Z
			MENUITEM "Mount as Image",				       ID_MOUNT_IMAGE_Z
			MENUITEM "Unmount",				       ID_UMOUNT_Z
		END
	END
END


// version resource
VS_VERSION_INFO VERSIONINFO
 FILEVERSION 0,74,0,0
 PRODUCTVERSION 0,74,0,0
 FILEFLAGSMASK 0x3fL
 FILEFLAGS 0x0L
 FILEOS 0x40004L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
		BLOCK "040904b0"
		BEGIN
		    VALUE "Comments", "(C) 2002-2014 DOSBox Team, published under GNU GPL"
		    VALUE "CompanyName", "DOSBox Team"
		    VALUE "FileDescription", "DOSBox DOS Emulator"
		    VALUE "FileVersion", "0, 74, 0, 0"
		    VALUE "InternalName", "DOSBox"
		    VALUE "LegalCopyright", "Copyright ?2002-2014 DOSBox Team"
		    VALUE "OriginalFilename", "dosbox.exe"
		    VALUE "ProductName", "DOSBox DOS Emulator"
		    VALUE "ProductVersion", "0, 74, 0, 0"
		END
    END
    BLOCK "VarFileInfo"
    BEGIN
		VALUE "Translation", 0x409, 1200
    END
END
