#!/bin/sh
git remote add github https://joncampbell123@github.com/joncampbell123/dosbox-x.git # might fail
git remote set-url github https://joncampbell123@github.com/joncampbell123/dosbox-x.git # might fail
git push github master

