
#ifndef __ISP_UTILS_V4_MISC_INFORMATIONAL_H
#define __ISP_UTILS_V4_MISC_INFORMATIONAL_H

/* informational #defines that may or may not have any effect on the data type,
 * but serve as help for the programmer */

/* Endianness reminders & markers.
 * And someday, when compiling for, say, PowerPC, it will map
 * to whatever native datatype typedefs that tell PowerPC what endianness to use */
#define _Little_Endian_
#define _Big_Endian_
#define _Mixed_Endian_

#endif /* __ISP_UTILS_V4_MISC_INFORMATIONAL_H */
