#!/bin/sh
./sb2u.pl cp437_to_unicode <CP437.TXT >../../include/cp437_uni.h || exit 1
./sb2u.pl cp737_to_unicode <CP737.TXT >../../include/cp737_uni.h || exit 1
./sb2u.pl cp775_to_unicode <CP775.TXT >../../include/cp775_uni.h || exit 1
./sb2u.pl cp808_to_unicode <CP808.TXT >../../include/cp808_uni.h || exit 1
./sb2u.pl cp850_to_unicode <CP850.TXT >../../include/cp850_uni.h || exit 1
./sb2u.pl cp852_to_unicode <CP852.TXT >../../include/cp852_uni.h || exit 1
./sb2u.pl cp853_to_unicode <CP853.TXT >../../include/cp853_uni.h || exit 1
./sb2u.pl cp855_to_unicode <CP855.TXT >../../include/cp855_uni.h || exit 1
./sb2u.pl cp856_to_unicode <CP856.TXT >../../include/cp856_uni.h || exit 1
./sb2u.pl cp857_to_unicode <CP857.TXT >../../include/cp857_uni.h || exit 1
./sb2u.pl cp858_to_unicode <CP858.TXT >../../include/cp858_uni.h || exit 1
./sb2u.pl cp859_to_unicode <CP859.TXT >../../include/cp859_uni.h || exit 1
./sb2u.pl cp860_to_unicode <CP860.TXT >../../include/cp860_uni.h || exit 1
./sb2u.pl cp861_to_unicode <CP861.TXT >../../include/cp861_uni.h || exit 1
./sb2u.pl cp862_to_unicode <CP862.TXT >../../include/cp862_uni.h || exit 1
./sb2u.pl cp863_to_unicode <CP863.TXT >../../include/cp863_uni.h || exit 1
./sb2u.pl cp864_to_unicode <CP864.TXT >../../include/cp864_uni.h || exit 1
./sb2u.pl cp865_to_unicode <CP865.TXT >../../include/cp865_uni.h || exit 1
./sb2u.pl cp866_to_unicode <CP866.TXT >../../include/cp866_uni.h || exit 1
./sb2u.pl cp867_to_unicode <CP867.TXT >../../include/cp867_uni.h || exit 1
./sb2u.pl cp868_to_unicode <CP868.TXT >../../include/cp868_uni.h || exit 1
./sb2u.pl cp869_to_unicode <CP869.TXT >../../include/cp869_uni.h || exit 1
./sb2u.pl cp872_to_unicode <CP872.TXT >../../include/cp872_uni.h || exit 1
./sb2u.pl cp874_to_unicode <CP874.TXT >../../include/cp874_uni.h || exit 1
./db2u.pl cp932_to_unicode <CP932.TXT >../../include/cp932_uni.h || exit 1
./db2u.pl cp936_to_unicode <CP936.TXT >../../include/cp936_uni.h || exit 1
./db2u.pl cp949_to_unicode <CP949.TXT >../../include/cp949_uni.h || exit 1
#./db2u.pl cp950_to_unicode <CP950.TXT >../../include/cp950_uni.h || exit 1 # For BIG5-1984 and BIG5-2003 with ChinaSea extension
#./db2u.pl cp951_to_unicode <CP951.TXT >../../include/cp951_uni.h || exit 1 # For BIG5-HKSCS and optionally BIG5 Unicode-At-On (UAO)
./sb2u.pl cp1250_to_unicode <CP1250.TXT >../../include/cp1250_uni.h || exit 1
./sb2u.pl cp1251_to_unicode <CP1251.TXT >../../include/cp1251_uni.h || exit 1
./sb2u.pl cp1252_to_unicode <CP1252.TXT >../../include/cp1252_uni.h || exit 1
./sb2u.pl cp1253_to_unicode <CP1253.TXT >../../include/cp1253_uni.h || exit 1
./sb2u.pl cp1254_to_unicode <CP1254.TXT >../../include/cp1254_uni.h || exit 1
./sb2u.pl cp1255_to_unicode <CP1255.TXT >../../include/cp1255_uni.h || exit 1
./sb2u.pl cp1256_to_unicode <CP1256.TXT >../../include/cp1256_uni.h || exit 1
./sb2u.pl cp1257_to_unicode <CP1257.TXT >../../include/cp1257_uni.h || exit 1
./sb2u.pl cp1258_to_unicode <CP1258.TXT >../../include/cp1258_uni.h || exit 1
./sb2u.pl cp3021_to_unicode <CP3021.TXT >../../include/cp3021_uni.h || exit 1
