
#include "winresrc.h"
#include "../include/resource.h"
#include "../vs/config_package.h"

#define VERSION_NUMBER 2026,06,02,0

// icon resource
dosbox_ico ICON "../contrib/icons/dosbox-x.ico"
IDI_MAPPER ICON "../contrib/icons/mapper.ico"
IDI_CFG_GUI ICON "../contrib/icons/cfggui.ico"
IDI_PAUSE ICON "../contrib/icons/pause.ico"

// version resource
VS_VERSION_INFO VERSIONINFO
 FILEVERSION VERSION_NUMBER
 PRODUCTVERSION VERSION_NUMBER
 FILEFLAGSMASK 0x3fL
 FILEFLAGS 0x0L
 FILEOS 0x40004L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
		BLOCK "040904b0"
		BEGIN
		    VALUE "Comments", "Project maintainer: joncampbell123 (TheGreatCodeholio)"
		    VALUE "CompanyName", "DOSBox-X Project"
		    VALUE "FileDescription", "DOSBox-X DOS Emulator"
		    VALUE "FileVersion", PACKAGE_VERSION
		    VALUE "InternalName", "DOSBox-X"
		    VALUE "LegalCopyright", "Copyright 2011-2022 The DOSBox-X Team (joncampbell123), published under GNU GPL"
		    VALUE "OriginalFilename", "dosbox-x.exe"
		    VALUE "ProductName", "DOSBox-X DOS Emulator"
		    VALUE "ProductVersion", PACKAGE_VERSION
		END
    END
    BLOCK "VarFileInfo"
    BEGIN
		VALUE "Translation", 0x409, 1200
    END
END
